"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.LiteralMacro = exports.CompilerMacro = exports.Macro = void 0;
const context_1 = require("./context");
const scan_1 = require("./scan");
const parse_1 = require("./parse");
const namespace_1 = require("./namespace");
const error = require("./error");
const value = require("./value");
const types = require("./datatypes");
const expr = require("./expr");
/**
 * Invokable block of code
 */
class Macro extends value.Value {
    constructor(token, type = null, recursive = false) {
        super(token, value.ValueType.Macro, null, type);
        this.type = value.ValueType.Macro;
        this.datatype = null;
        /**
         * Datatypes which the macro satisfies
         */
        // private matchingDatatypes: types.ArrowType[] = [];
        // private failedDatatypes: types.ArrowType[] = [];
        /**
         * Did the user flag this macro as recursive?
         */
        this.recursive = false;
        this.recursive = recursive;
    }
    /**
     * Determine the complete datatype of the macro given only input types
     * @param ctx context
     * @param token location in the code where inferrence is needed
     * @param inputTypes required given input datatypes
     * @returns An arrow type for the datatype of the macro
     */
    inferDatatype(ctx, inputTypes, token = this.token) {
        // const cached = this.matchingDatatypes.find(t => t.checkInputTypes(inputTypes));
        // if (cached)
        //     return cached;
        // Generate dummy inputs
        const inputs = inputTypes.map(t => expr.DummyDataExpr.create(token, t));
        ctx.stack.push(...inputs);
        // Trace the macro
        const ios = ctx.traceIO(this, token);
        if (ios instanceof error.SyntaxError)
            return ios;
        ctx.popn(inputs.length);
        // Validate trace
        if (ios.takes.length > inputs.length)
            return 'differing input lengths';
        if (ios.takes.some((e, i) => e !== inputs[i])) {
            console.log(ios.takes, 'vs', inputs);
            return 'differing input values';
        }
        if (ios.takes.some(v => !v.datatype))
            throw new Error('wtf?');
        if (ios.gives.some(v => !v.datatype))
            return new error.SyntaxError(`macro returns untyped value ${ios.gives.find(v => !v.datatype).typename()}`, token, ctx);
        // Add match
        // TODO the outputTypes should be merged with unused inputtypes
        const ret = new types.ArrowType(token, inputTypes, ios.gives.map(v => v.datatype));
        return ret;
    }
    /**
     * Verify this macro has specified type
     * @param ctx parser context
     * @param datatype type to check against
     * @param token location in code where check is required
     * @returns true/false if checks or not or a syntax error
     */
    typeCheck(ctx, datatype, token = this.token, safe = false) {
        // No need to inference if we already know the datatype
        if (this.datatype)
            return datatype.check(this.datatype);
        // if (this.matchingDatatypes.some(t => t.check(datatype)))
        //     return true;
        // if (this.failedDatatypes.some(t => t.check(datatype)))
        //     return false;
        // Generate type from partial
        const baseType = datatype.getBaseType();
        const dt = this.inferDatatype(ctx, baseType instanceof types.TupleType
            ? baseType.types
            : baseType.inputTypes, token);
        // Invalid macro
        // TODO this could mean it's just incorrectly typed...
        if (dt instanceof error.SyntaxError)
            return !safe && dt;
        // Verify type
        if (dt instanceof types.ArrowType) {
            const ret = datatype.check(this.datatype);
            // if (ret)
            //     this.matchingDatatypes.push(dt);
            // else
            //     this.failedDatatypes.push(dt);
            return ret;
        }
        // Invalid type
        // TODO shouldn't need these warnings in the future
        ctx.warn(token, 'could not infer partial type: ' + (dt || ''));
        return false;
        // return new error.SyntaxError('could not infer partial type: ' + (dt || ''), token, ctx);
    }
}
exports.Macro = Macro;
/**
 * A macro that is created internally by the compiler
 */
class CompilerMacro extends Macro {
    /**
     * @param invokeAction - body of the macro
     * @param [name] - debugging symbol for the macro
     */
    constructor(token, invokeAction, name, datatype) {
        super(token, datatype);
        this.invokeAction = invokeAction;
        this.name = name;
    }
    /**
     * @override
     */
    action(ctx, token) {
        return this.invokeAction(ctx, token);
    }
    // TODO toString for debugging
    toString() {
        return `CompilerMacro { ${this.name} }`;
    }
}
exports.CompilerMacro = CompilerMacro;
/**
 * User-defined macros
 */
class LiteralMacro extends Macro {
    /**
     * Construct Macro object from literal token
     * @param ctx - context for literal
     * @param token - token for literal
     */
    constructor(ctx, token) {
        super(token);
        this.body = token.body;
        this.scopes = ctx.scopes.slice();
        this.recursive = token instanceof scan_1.MacroToken && token.recursive;
    }
    /**
     * @override
     */
    action(ctx) {
        // TODO simplify and/or use ctx.copyState()
        // Use proper lexical scope
        const oldScopes = ctx.scopes;
        ctx.scopes = this.scopes;
        ctx.scopes.push({});
        // Invoke body
        let ret;
        try {
            ret = (0, parse_1.default)(this.body, ctx);
        }
        catch (e) {
            // Always Restore ctx state
            ctx.scopes.pop();
            ctx.scopes = oldScopes;
            throw e;
        }
        // Restore ctx state
        ctx.scopes.pop();
        ctx.scopes = oldScopes;
        return ret;
    }
    /**
     * Handle namespace call
     * @param ctx - context object
     * @param token - invokee token
     * @returns - on success return namespace accessor macro on otherwise returns error
     */
    getNamespace(ctx, token) {
        // TODO simplify and/or use ctx.copyState()
        // Use proper lexical scope
        const oldScopes = ctx.scopes;
        ctx.scopes = this.scopes;
        ctx.scopes.push({});
        // Invoke body
        let ret;
        try {
            ret = ctx.toError((0, parse_1.default)(this.body, ctx), token);
        }
        catch (e) {
            // Always Restore ctx state
            ctx.scopes.pop();
            ctx.scopes = oldScopes;
            throw e;
        }
        // Restore ctx state
        const newScope = ctx.scopes.pop();
        ctx.scopes = oldScopes;
        // On successs return the scope otherwise give the error
        return ret instanceof context_1.default
            ? new value.NamespaceValue(token, new namespace_1.Namespace(newScope, token))
            : ret;
    }
    /**
     *
     * @param ctx compiler context
     * @param inputs input datatypes
     * @param outputs output datatypes
     * @returns void if successful, anything else if error
     */
    applyType(ctx, inputs, outputs) {
        // Infer output types from inputs
        const type = this.inferDatatype(ctx, inputs.types, this.token);
        if (typeof type == 'string')
            return new error.SyntaxError(type, this.token, ctx);
        if (type instanceof error.SyntaxError)
            return type;
        // Verify outputs
        if (outputs && !outputs.types.every((t, i) => t.check(type.outputTypes[i]))) {
            ctx.warn(this.token, 'incorrectly typed macro');
            console.warn('incorrectly typed macro, should be', type);
            // return type;
        }
        // Apply datatype
        this.datatype = type;
    }
    toString() {
        return `LiteralMacro { ${this.token.file || ''}:${this.token.position} }`;
    }
}
exports.LiteralMacro = LiteralMacro;
//# sourceMappingURL=macro.js.map