"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.generatePerfSummary = exports.runtimeCache = void 0;
const scan_1 = require("./scan");
const value = require("./value");
const context_1 = require("./context");
const macro_1 = require("./macro");
const error = require("./error");
const types = require("./datatypes");
/*
The name for this file is somewhat misleading but technically correct

Because lang is postfix, just having a lexer is enough to create AST, however
For this language user interacts with this module through operators in order to
make a somewhat different expression tree
*/
exports.runtimeCache = [];
function generatePerfSummary() {
    const times = {};
    exports.runtimeCache.sort(([, ta], [, tb]) => tb - ta);
    exports.runtimeCache.forEach(([t, v]) => {
        if (times[t])
            times[t] += v;
        else
            times[t] = v;
    });
    return {
        times: exports.runtimeCache.splice(0),
        uniqueTimes: times,
    };
}
exports.generatePerfSummary = generatePerfSummary;
/**
 * Create parse tree
 *
 * @param tokens - tokens to parse
 * @param ctx - parse ctx
 */
function parse(tokens, ctx = new context_1.default(undefined, tokens[0].file)) {
    // For each token
    for (let i = 0; i < tokens.length; i++) {
        const t = tokens[i];
        switch (t.type) {
            // Identifiers: references to values
            case scan_1.LexerToken.Type.Identifier:
                // Handle subtypes
                if (t.isEscaped) {
                    // Escaped symbol
                    ctx.push(new value.IdValue(t));
                }
                else {
                    // Invoke operator
                    // const s = performance.now();
                    const v = ctx.getId(t.value);
                    if (!v)
                        return new error.SyntaxError(`${t.token} is undefined`, t, ctx);
                    // console.log('invoke', t.token, v);
                    const ret = ctx.invoke(v, t);
                    // runtimeCache.push([t.token, performance.now() - s]);
                    // TODO this returns null sometimes
                    if (!(ret instanceof context_1.default))
                        return ret;
                }
                break;
            // Need to determine number type from literal
            case scan_1.LexerToken.Type.Number:
                ctx.push(new value.NumberValue(t, t.value));
                break;
            // Blocks: Need to form a closure with current scope
            case scan_1.LexerToken.Type.Block: {
                // This shouldn't be needed
                if (!(t instanceof scan_1.MacroToken))
                    throw new Error('wtf?');
                // Get input and output types
                const ts = [];
                for (const tt of t.types) {
                    const rv = ctx.parseTuple(tt, true);
                    if (!(rv instanceof context_1.default))
                        return rv;
                    const v = ctx.pop();
                    if (!(v.value instanceof types.TupleType))
                        return new error.SyntaxError('Macro inputs tuple should only contain types', tt, ctx);
                    ts.push(v.value);
                }
                // Push macro onto stack
                const m = new macro_1.LiteralMacro(ctx, t);
                m.recursive = t.recursive;
                if (ts.length !== 0 && !m.recursive) {
                    const rv = m.applyType(ctx, ts[0], ts[1]);
                    if (rv instanceof types.ArrowType)
                        console.warn('inferred type: ', rv);
                    if (rv)
                        throw rv;
                }
                ctx.push(m);
                break;
            }
            // Tuples: Like a block but gets parsed immediately and not as smart
            case scan_1.LexerToken.Type.Tuple: {
                const ret = ctx.parseTuple(t);
                if (!(ret instanceof context_1.default))
                    return ret;
                break;
            }
            // Parse: strings
            case scan_1.LexerToken.Type.String:
                ctx.push(new value.StrValue(t));
                break;
        }
    }
    return ctx;
}
exports.default = parse;
//# sourceMappingURL=parse.js.map